@@ -1,5 +1,6 @@
 from __future__ import absolute_import
 
+import time
 import string
 import xapian
 
@@ -147,6 +148,11 @@ class Database(object):
 
     :param spelling: If True, write spelling correction data to the
     database.
+
+    :param replicated: If True, the database is opened read-only in
+    replication mode (reopen() is never called, see xapian ticket
+    #434)
+
     """
 
     record_factory = record_factory
@@ -158,24 +164,33 @@ class Database(object):
     value_count_name = "_XODB_COUNT_"
     backend = None
 
-    def __init__(self, path, writable=True, overwrite=False, spelling=True):
+    def __init__(self, path,
+                 writable=True,
+                 overwrite=False,
+                 spelling=True,
+                 replicated=False):
         self.db_path = path
         self._writable = writable
         self._overwrite = overwrite
         self.spelling = spelling
-        self.open()
+        self.replicated = replicated
 
-    def open(self):
-        if isinstance(self.db_path, basestring):
-            if self._writable:
-                if self._overwrite:
+        if isinstance(path, basestring):
+            if writable:
+                if replicated:
+                    raise TypeError("replication cannot be used with "
+                                    "writable databases")
+                if overwrite:
                     flags = xapian.DB_CREATE_OR_OVERWRITE
                 else:
                     flags = xapian.DB_CREATE_OR_OPEN
-                self.backend = xapian.WritableDatabase(self.db_path, flags)
+                self.backend = xapian.WritableDatabase(path, flags)
             else:
-                self.backend = xapian.Database(self.db_path)
-        elif isinstance(self.db_path, xapian.Database):
+                self.backend = xapian.Database(path)
+        elif isinstance(path, xapian.Database):
+            if replicated:
+                raise TypeError("replication can only be used "
+                                "if a database path is provided.")
             self.backend = self.db_path
 
         self.relevance_prefixes = {}
@@ -183,11 +198,14 @@ class Database(object):
         self.values = {}
         self.value_sorts = {}
         self.type_map = {}
-        self.reopen()
+        self.refresh_metadata()
 
     def close(self):
         self.backend.close()
 
+    def flush(self):
+        self.backend.flush()
+
     def map(self, otype, schema):
         """Map a type to a schema."""
         self.type_map[otype] = schema
@@ -289,36 +307,49 @@ class Database(object):
                 self.backend.add_document(obj)
                 added.append(obj)
                 continue
-
-            if hasattr(obj, '__xodb_schema__'):
-                schema_type = obj.__xodb_schema__
-            else:
-                schema_type = self.schema_for(type(obj))
-            schema = schema_type.from_defaults()
-            schema.update_by_object(obj)
-
-            if validate and not schema.validate():
-                invalid = []
-                for child in schema.all_children:
-                    if not child.valid:
-                        invalid.append(child)
-                raise ValidationError("Elements of %s did not validate %s:" %
-                                      (schema.__class__.__name__,
-                                       list((c.name, c.value)
-                                            for c in invalid)))
-
-            doc = schema.to_document(self)
-            data = dumps((_schema_name(schema_type), schema.flatten()))
-            doc.set_data(data)
+            doc = self.to_document(obj, validate=validate)
             self.backend.add_document(doc)
             added.append(doc)
         return added
 
+    def to_document(self, obj, validate=True):
+        if hasattr(obj, '__xodb_schema__'):
+            schema_type = obj.__xodb_schema__
+        else:
+            schema_type = self.schema_for(type(obj))
+        schema = schema_type.from_defaults()
+        schema.update_by_object(obj)
+
+        if validate and not schema.validate():
+            invalid = []
+            for child in schema.all_children:
+                if not child.valid:
+                    invalid.append(child)
+            raise ValidationError("Elements of %s did not validate %s:" %
+                                  (schema.__class__.__name__,
+                                   list((c.name, c.value)
+                                        for c in invalid)))
+        doc = schema.to_document(self)
+        data = dumps((_schema_name(schema_type), schema.flatten()))
+        doc.set_data(data)
+        return doc
+
     def reopen(self):
+        if not self.replicated:
+            self.backend.reopen()
+        else:
+            # replication does not support the reopen() method, so the
+            # db must be explicitly closed and reopened.
+            assert self.db_path, ("Must provide a db path when "
+                                  "using replication.")
+            self.close()
+            self.backend = xapian.Database(self.db_path)
+        self.refresh_metadata()
+
+    def refresh_metadata(self):
         """Reopen the database, refresing the local connection's view
         of the prefix and value data.
         """
-        self.backend.reopen()
         try:
             keys = self.backend.metadata_keys()
         except UnimplementedError:
@@ -396,8 +427,11 @@ class Database(object):
             qp.set_stemming_strategy(QueryParser.STEM_SOME)
         return qp
 
-    def querify(self, query, language=None, translit='translit/long',
-                default_op=Query.OP_AND, parser_flags=default_parser_flags):
+    def querify(self, query,
+                language=None,
+                translit='translit/long',
+                default_op=Query.OP_AND,
+                parser_flags=default_parser_flags):
         """Return a query object, constructed from a string, query
         object, or mixed sequence of any number of strings, query
         objects, or subsequences of the same form.
@@ -428,10 +462,19 @@ class Database(object):
                                         default_op,
                                         parser_flags) for q in query))
 
-    def query(self, query, offset=0, limit=0, order=None,
-              reverse=False, language=None, check=0, translit='translit/long',
-              match_decider=None, match_spy=None, document=False,
-              echo=False, parser_flags=default_parser_flags,
+    def query(self, query,
+              offset=0,
+              limit=0,
+              order=None,
+              reverse=False,
+              language=None,
+              check=0,
+              translit='translit/long',
+              match_decider=None,
+              match_spy=None,
+              document=False,
+              echo=False,
+              parser_flags=default_parser_flags,
               default_op=Query.OP_AND):
         """
         Query the database with the provided string or xapian Query
@@ -441,19 +484,29 @@ class Database(object):
         self.reopen()
 
         enq = xapian.Enquire(self.backend)
+        if echo:
+            print "Parsing query."
         query = self.querify(query, language, translit,
                              default_op, parser_flags)
         if echo:
-            print str(query)
+            print "Done parsing query: %s" % str(query)
         enq.set_query(query)
 
         limit = limit or self.backend.get_doccount()
 
+        if echo:
+            start = time.time()
+            print "Fetching mset..."
         mset = self._build_mset(enq, offset, limit, order, reverse,
                                 check, match_decider, match_spy)
+        if echo:
+            print "Fetched mset in %s" % str(time.time() - start)
         return self._return_objects(mset, document=document)
 
-    def count(self, query="", language=None, echo=False,
+    def count(self,
+              query="",
+              language=None,
+              echo=False,
               translit='translit/long',
               parser_flags=default_parser_flags,
               default_op=Query.OP_AND):
@@ -539,9 +592,14 @@ class Database(object):
                 results[(name, score)] = score
         return results
 
-    def estimate(self, query, limit=0, klimit=1.0, language=None,
+    def estimate(self, query,
+                 limit=0,
+                 klimit=1.0,
+                 language=None,
                  translit='translit/long',
-                 default_op=Query.OP_AND, parser_flags=default_parser_flags):
+                 check_at_least=None,
+                 default_op=Query.OP_AND,
+                 parser_flags=default_parser_flags):
         """Estimate the number of documents that will be yielded with the
         given query.
 
@@ -553,10 +611,16 @@ class Database(object):
         if limit == 0:
             limit = int(self.backend.get_doccount() * klimit)
 
-        query = self.querify(query, language, translit, default_op, parser_flags)
+        query = self.querify(query, language, translit,
+                             default_op, parser_flags)
 
         enq.set_query(query)
-        return enq.get_mset(0, 0, limit).get_matches_estimated()
+        if check_at_least is not None:
+            mset = enq.get_mset(0, 0, limit)
+        else:
+            mset = enq.get_mset(0, 0, limit)
+            
+        return mset.get_matches_estimated()
 
     def term_freq(self, term):
         """
@@ -566,8 +630,9 @@ class Database(object):
         self.backend.reopen()
         return self.backend.get_termfreq(term)
 
-    def describe_query(self, query, language=None,
-              default_op=Query.OP_AND):
+    def describe_query(self, query,
+                       language=None,
+                       default_op=Query.OP_AND):
         """
         Describe the parsed query.
         """
@@ -575,7 +640,8 @@ class Database(object):
         q = qp.parse_query(query, default_parser_flags)
         return str(q)
 
-    def spell(self, query, language=None):
+    def spell(self, query,
+              language=None):
         """
         Suggest a query string with corrected spelling.
         """
@@ -584,10 +650,20 @@ class Database(object):
         qp.parse_query(query, QueryParser.FLAG_SPELLING_CORRECTION)
         return qp.get_corrected_query_string().decode('utf8')
 
-    def suggest(self, query, offset=0, limit=0, moffset=0, mlimit=0,
-                klimit=1.0, kmlimit=1.0, translit='translit/long',
-                language=None, prefix=None, decider=None, score=False,
-                echo=False, default_op=Query.OP_AND,
+    def suggest(self, query,
+                offset=0,
+                limit=0,
+                moffset=0,
+                mlimit=0,
+                klimit=1.0,
+                kmlimit=1.0,
+                translit='translit/long',
+                language=None,
+                prefix=None,
+                decider=None,
+                score=False,
+                echo=False,
+                default_op=Query.OP_AND,
                 parser_flags=default_parser_flags):
         """
         Suggest terms that would possibly yield more relevant results
@@ -640,8 +716,14 @@ class Database(object):
             else:
                 yield val
 
-    def _build_mset(self, enq, offset=0, limit=None, order=None, reverse=False,
-                    check=None, match_decider=None, match_spy=None):
+    def _build_mset(self, enq,
+                    offset=0,
+                    limit=None,
+                    order=None,
+                    reverse=False,
+                    check=None,
+                    match_decider=None,
+                    match_spy=None):
         if order is not None:
             if isinstance(order, basestring):
                 try:
