@@ -18,6 +18,9 @@ MAX_TERM_LEN = 240 # compile time xapian limit
 def _prefix(name):
     return (u'X%s:' % name.upper()).encode('utf-8')
 
+def _normalize(value):
+    return unicodedata.normalize(
+        'NFKC', unicode(value).strip().lower()).encode('utf-8')
 
 _use_schema = object() # marker says use schema term generator
 
@@ -189,8 +192,6 @@ class Schema(schema.Form):
         if element.translit:
             value = value.encode(element.translit)
 
-        value = value.encode('utf-8')
-        
         if element.string:
             if element.string_prefix:
                 self._add_term(value, element.string_prefix, element.boolean)
@@ -200,6 +201,7 @@ class Schema(schema.Form):
         if element.sortable:
             self._add_value(value, name, 'string')
 
+        value = _normalize(value)
         if element.prefix:
             self._check_prefix(name, element.boolean)
             index_text(value, 1, _prefix(name))
@@ -255,8 +257,7 @@ class Schema(schema.Form):
         return tg
 
     def _add_term(self, value, prefix=None, boolean=False):
-        value = unicodedata.normalize(
-            'NFKC', unicode(value).strip().lower()).encode('utf-8')
+        value = _normalize(value)
         if len(value) > MAX_TERM_LEN:
             raise InvalidTermError('The term %s is too long.' % value)
         if prefix:
